#include "femPost.h"
#include "femSolver.h"
#include "femProblem.h"
#include "femGmsh.h"
#include "femGroup.h"
#include "femEntity.h"
#include <numeric>
#include <fstream>

namespace fem
{

Post::Post(Solver &_solver) : verbose(false), solver(_solver)
{
    // global gmsh options
    gmsh::option::setNumber("General.Orthographic", 0); // better perspective
    gmsh::option::setNumber("Mesh.VolumeEdges", 0);     // don't display the mesh
    gmsh::option::setNumber("Mesh.SurfaceEdges", 0);    // don't display the mesh
}

Post::~Post()
{
    if (debug_level() > 0)
        std::cout << "~Post()\n";
}

/// rebuild solution vectors for gmsh and add views

void Post::build_views()
{
    std::cout << "creating views...\n";
    // nodal dofs
    add_dofs_view(solver.pbl, solver.dofs, solver.X, solver.dofPart, solver.dofIdx);

    // nodal vectors
    add_nodalvector_view(solver.pbl, "displacement", solver.dofs, solver.X, solver.dofPart, solver.dofIdx);
    add_nodalvector_view(solver.pbl, "force", solver.dofs, solver.f, solver.dofPart, solver.dofIdx);

    // compute heat fluxes, strains, stresses

    std::vector<size_t> etags_all;                              // tags of elements
    std::vector<std::vector<double>> sigma_all;                 // values of stresses at the nodes of each element
    std::vector<std::vector<double>> epsilon_all;               // values of strains at the nodes of each element
    std::map<size_t, std::vector<Eigen::VectorXd>> sig_by_node; // value of all the stress vectors (Voigt) at a given node (key of the dict)
    std::map<size_t, std::vector<Eigen::VectorXd>> eps_by_node; // same for strain vector
    solver.compute_sig_epl(etags_all, sigma_all, epsilon_all, sig_by_node, eps_by_node);

    // ...discontinuous by node by element
    add_tensor_views("stress", etags_all, sigma_all);
    add_tensor_views("strain", etags_all, epsilon_all);

    // ...continuous averaged nodal fields
    add_smooth_views("stress", sig_by_node);
    add_smooth_views("strain", eps_by_node);
}

/// add 1 scalar view for each dof type to gmsh.

void Post::add_dofs_view(Problem &pbl, DOFs &dofs,
                         std::vector<Eigen::VectorXd *> &X,
                         Eigen::ArrayXXi &dofPart,
                         Eigen::ArrayXXi &dofIdx)
{
    // scalar view of each dof
    for (size_t di = 0; di < dofs.size(); ++di) // TODO: iterator
    {
        DOF dof = dofs[di]; // extract each dof type separately
        std::vector<double> x_all(pbl.nodeTags.size(), 0.0);
        for (size_t i = 0; i < pbl.nodeTags.size(); ++i)
        {
            size_t ntag = pbl.nodeTags[i];
            size_t gidx = pbl.nodemap[ntag];
            int part = dofPart(gidx, dof.index());
            int idx = dofIdx(gidx, dof.index());
            x_all[i] = (*X[part])[idx];
        }
        // Export solution to gmsh
        std::string name = dofs.name(dof);
        int sview = add_view(name);
        gmsh::view::addHomogeneousModelData(sview, 0, "", "NodeData",
                                            pbl.nodeTags, x_all);
    }
}

// add 1 tensor view to gmsh and 6 scalar views for the components.
// tensor_all is in gmsh format: array of "nbelem" elements (same size as "etags_all")
// each element contains the tensor data:
// [a11 a12 a13 a21 a22 a23 a31 a32 a33 a11 a12 a13 a21 a22 a23 a31 a32 a33 ...]
// <--------------node1--------------> <--------------node2--------------> <--

void Post::add_tensor_views(std::string const &vname,
                            std::vector<size_t> const &etags_all,
                            std::vector<std::vector<double>> const &tensor_all)
{
    // std::cout << "Post::add_tensor_views " << vname << "\n";
    // std::cout << "\tetags_all.size() = " << etags_all.size() << "\n";
    // std::cout << "\ttensor_all.size() = " << tensor_all.size() << "\n";

    // tensor view (by node, by element)
    std::string name = vname + "_tensor";
    int sview = add_view(name);
    gmsh::view::addModelData(sview, 0, "", "ElementNodeData",
                             etags_all, tensor_all, 0.0, 9);

    // component view (by node, by element)
    //   components are retrieved from the tensor data in gmsh format
    // [ [0]     [ [XX]     [ [0] [1] [2]    [ [XX] [XY] [XZ]
    //   [4]       [YY]       [3] [4] [5]      [XY] [YY] [YZ]
    //   [8]       [ZZ]       [6] [7] [8] ]    [XZ] [YZ] [ZZ] ]
    //   [1or3]    [XY]
    //   [2or6]    [XZ]
    //   [5or7] ]  [YZ] ]

    std::vector<int> comp_id = {0, 4, 8, 1, 2, 5};
    std::vector<std::string> comp_name = {"xx", "yy", "zz", "xy", "xz", "yz"};
    for (size_t i = 0; i < comp_id.size(); ++i)
    {
        std::string name = vname + "_" + comp_name[i];
        int sview = add_view(name);
        std::vector<std::vector<double>> comp(tensor_all.size());
        for (size_t ie = 0; ie < tensor_all.size(); ++ie)
        {
            comp[ie].resize(tensor_all[ie].size() / 9);
            for (size_t in = 0; in < tensor_all[ie].size() / 9; ++in)
                comp[ie][in] = tensor_all[ie][9 * in + comp_id[i]];
        }
        gmsh::view::addModelData(sview, 0, "", "ElementNodeData",
                                 etags_all, comp, 0.0, 1);
    }
}

/// add the nodal vector (displacement, forces) as a new gmsh view.

void Post::add_nodalvector_view(Problem &pbl,
                                std::string const &vname,
                                DOFs &dofs,
                                std::vector<Eigen::VectorXd *> &X,
                                Eigen::ArrayXXi &dofPart,
                                Eigen::ArrayXXi &dofIdx)
{
    std::vector<std::vector<double>> uvw_all(pbl.nodeTags.size());
    for (size_t i = 0; i < pbl.nodeTags.size(); ++i)
    {
        uvw_all[i].resize(3);
        size_t ntag = pbl.nodeTags[i];
        size_t gidx = pbl.nodemap[ntag];
        for (size_t di = 0; di < dofs.size(); ++di)
        {
            DOF dof = dofs[di];
            int part = dofPart(gidx, dof.index());
            int idx = dofIdx(gidx, dof.index());
            uvw_all[i][di] = (*X[part])[idx];
        }
    }
    // Export solution to gmsh
    std::string name = vname + "_vector";
    int sview = add_view(name);
    gmsh::view::addModelData(sview, 0, "", "NodeData",
                             pbl.nodeTags, uvw_all);
}

/// add 6 scalar views for the components of a smoothed field (stresses or strains)
///         voigt_by_node is a map such that:
///         voigt_by_node[nodetag] = array of vectors of stresses/strains (Voigt - 6 comps)
///                                  evaluated from the neighbouring elements

void Post::add_smooth_views(std::string const &vname,
                            std::map<size_t, std::vector<Eigen::VectorXd>> &voigt_by_node)
{
    std::vector<Eigen::VectorXd> smooth_sig;
    std::vector<size_t> smooth_ntags;

    std::vector<std::vector<double>> tensors;

    // compute the average

    for (auto const &[ntag, sigs] : voigt_by_node)
    {
        // auto s = std::reduce(sigs.begin(), sigs.end()) / sigs.size(); // C++17 / not avail in gcc 8.1 mingw
        Eigen::VectorXd s0 = Eigen::VectorXd::Zero(6);
        // std::cout << "ntag=" << ntag << " sigs.size() = " << sigs.size() << std::endl;
        //  auto s = std::accumulate(sigs.begin(), sigs.end(), s0) / sigs.size(); // core dump sous gcc 9.4 ubuntu ???
        for (auto &v : sigs)
            s0 += v;
        auto s = s0 / sigs.size();
        // std::cout << "s=" << s << std::endl;
        smooth_sig.push_back(s);
        smooth_ntags.push_back(ntag);

        tensors.push_back({s[0], s[3], s[4], s[3], s[1], s[5], s[4], s[5], s[2]});
    }

    // tensor view (by node)
    std::string name = "smooth_" + vname + "_tensor";
    int sview = add_view(name);
    gmsh::view::addModelData(sview, 0, "", "NodeData",
                             smooth_ntags, tensors, 0.0, 9);

    // loop over the 6 components

    std::vector<std::string> comps = {"xx", "yy", "zz", "xy", "xz", "yz"};

    for (size_t c = 0; c < comps.size(); ++c)
    {
        std::string name = "smooth_" + vname + "_" + comps[c];
        int sview = add_view(name);
        std::vector<double> comp;
        for (auto &s : smooth_sig)
            comp.push_back(s[c]);
        gmsh::view::addHomogeneousModelData(sview, 0, "", "NodeData",
                                            smooth_ntags, comp);
    }

    // TODO: we could rebuild the average tensor here [DONE]
    //       we could also calculate the 2nd invariant (Von Mises)
}

/// add a new gmsh view and configure it
/// store the view in a dict called 'views' so that the views can be easily configured
/// later by the user using their names

int Post::add_view(std::string const &name)
{
    // if(self.verbose): print(f'\tadding new view: {name}')

    int sview = gmsh::view::add(name);
    // sview = 1 for the first view although it is named Views[0] in the GUI and the options!!
    views[name] = sview;
    configure(sview);
    //set_viridis(sview); // <= already present in gmsh with 'F4' key (View.ColormapNumber=23)
    return sview;
}

void Post::set_camera_3D()
{
    gmsh::option::setNumber("General.TrackballQuaternion0", 0.62);
    gmsh::option::setNumber("General.TrackballQuaternion1", 0.17);
    gmsh::option::setNumber("General.TrackballQuaternion2", 0.20);
    gmsh::option::setNumber("General.TrackballQuaternion3", 0.74);
}

void Post::deform(double factor)
{
    int v_disp = views["displacement_vector"];
    for (auto const &[vname, v] : views)
    {
        if (v != v_disp)
        {
            gmsh::option::setNumber("View[" + std::to_string(v - 1) + "].UseGeneralizedRaise", 1);
            gmsh::option::setNumber("View[" + std::to_string(v - 1) + "].GeneralizedRaiseView", v_disp - 1);
            gmsh::option::setNumber("View[" + std::to_string(v - 1) + "].GeneralizedRaiseFactor", factor);
        }
    }

    // v_disp = self.views['displacement_vector']
    // for vname in self.views:
    //     v = self.views[vname]
    //     if v != v_disp:
    //         gmsh.option.setNumber(f"View[{v-1}].UseGeneralizedRaise", 1)
    //         gmsh.option.setNumber(f"View[{v-1}].GeneralizedRaiseView", v_disp-1)
    //         gmsh.option.setNumber(f"View[{v-1}].GeneralizedRaiseFactor", factor)
}

void Post::configure(int sview)
{
    // scalar views
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].IntervalsType", 3); // filled iso
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].NbIso", 20);        //

    // vector views
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].GlyphLocation", 2); // at the node (default=barycenter)
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].VectorType", 2);    // simple Arrow (default=3D arrow)

    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].ColormapNumber", 23);  // Viridis colormap   
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].LightLines", 0);  // disable lighting for lines
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].LightTwoSide", 0);  // disable 2-sided lighting
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].ShowElement", 1);  // view black mesh outline
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].DrawSkinOnly", 1); // display skin only (should be faster)
    gmsh::option::setNumber("View[" + std::to_string(sview - 1) + "].Visible", 0);      // hide all views
}

/// display results

void Post::view()
{
    std::cout << "viewing results...\n";
    gmsh::fltk::run();
}

void Post::show_views(std::vector<std::string> const &names)
{
    for (auto &v : names)
    {
        auto it = views.find(v);
        if (it != views.end())
            gmsh::option::setNumber("View[" + std::to_string(it->second - 1) + "].Visible", 1);
        else
            std::cout << "WARNING: view named \"" << v << "\" does not exist!\n";
    }
}

/// write mesh/data to disk

void Post::write()
{
    gmsh::option::setNumber("Mesh.Binary", 1); // writes files in binary format

    gmsh::write("mesh.opt"); //  save the options
    gmsh::write("mesh.msh"); //  only save the mesh

    // save the views (without the mesh)
    gmsh::option::setNumber("PostProcessing.SaveMesh", 0);
    gmsh::option::setNumber("PostProcessing.SaveInterpolationMatrices", 0);

    for (auto const &[name, sview] : views)
    {
        // gmsh::view::write(sview, name+".pos"); // pos format: list-based data (indep. of any model)
        gmsh::view::write(sview, name + ".msh"); // writes the mesh in each file - each view is in a separate file
        // gmsh::view::write(sview, "mesh.msh", true); // single file, but the mesh is repeated
    }
}

std::vector<double>
Post::probe(std::string const &name, std::string const &grpname)
{
    // std::cout << "Post::probe(" << grpname << ")\n";
    // get physical group from grpname
    Group *grp = solver.pbl.groups_by_name.at(grpname); // TODO: better handling of errors

    // get the view
    auto it = views.find(name);
    if (it == views.end())
        throw std::runtime_error("WARNING: view named \"" + name + "\" does not exist!");
    int nview = it->second;

    // probe values
    std::vector<double> all_values;

    std::vector<std::size_t> nodeTags;
    std::vector<double> coord;
    gmsh::model::mesh::getNodesForPhysicalGroup(grp->dim, grp->tag, nodeTags, coord);
    // j'ai essayé de boucler sur les entités du groupe et faire un getNode de l'entité 
    // mais ça ne me retourne rien (?)

    // std::cout << "nodeTags.size()=" << nodeTags.size() << '\n';

    for (size_t i = 0; i < nodeTags.size(); ++i)
    {
        std::vector<double> values;
        double distance = 0.0;
        int step = -1;
        int numComp = -1; // do not use numComp=comp (gmsh returns [ ] for comp!=0 )
        bool gradient = false;
        double distanceMax = -1.; // find closest node
        gmsh::view::probe(nview, coord[i * 3], coord[i * 3 + 1], coord[i * 3 + 2], values,
                          distance, step, numComp, gradient, distanceMax);

        // debug
        // std::cout << "probing " << name << " at node #" << nodeTags[i] << ":";
        // std::cout << "\tnview: " << nview << '\n';
        // std::cout << "\tcoords: " << coord[i * 3] << ", " << coord[i * 3 + 1] << ", " << coord[i * 3 + 2] << '\n';
        // std::cout << "\tgot from gmsh: " << values << '\n';
        // std::cout << "\tdistance = " << distance << std::endl;

        all_values.insert(all_values.end(), values.begin(), values.end());
    }

    return all_values;
}

/// Probe a field at a given nodetag

std::vector<double>
Post::probe(std::string const &name, size_t nodetag)
{
    // get the view
    auto it = views.find(name);
    if (it == views.end())
        throw std::runtime_error("WARNING: view named \"" + name + "\" does not exist!");
    int nview = it->second;

    // retrieve node coordinates
    std::vector<double> coord;
    std::vector<double> parametricCoord; // unused
    int dim;                             // unused
    int tag;                             // unused
    gmsh::model::mesh::getNode(nodetag, coord, parametricCoord, dim, tag);

    // probe value

    std::vector<double> values;
    double distance = 0.0;
    int step = -1;
    int numComp = -1; // do not use numComp=comp (gmsh returns [ ] for comp!=0 )
    bool gradient = false;
    double distanceMax = -1.; // find closest node
    gmsh::view::probe(nview, coord[0], coord[1], coord[2], values, distance, step, numComp, gradient, distanceMax);

    // debug
    // std::cout << "probing " << name << " at node #" << nodetag << ":";
    // std::cout << "\tnview: "<< nview << '\n';
    // std::cout << "\tcoords: "<< coord << '\n';
    // std::cout << "\tgot from gmsh: "<< values << '\n';
    // std::cout << "\tdistance = "<< distance << std::endl;

    // std::cout << "probing \"" << name << "\" (comp=" << comp
    //           << ") at node #" << nodetag << ": " << values[comp] << std::endl;

    return values;
}

FEM_API std::ostream &
operator<<(std::ostream &out, Post const &obj)
{
    out << "Post\n";
    return out;
}


} // namespace fem
